#include "othello.h"

/* H.Takeshima's thinking routine Ver 2.1(98/05/05) */

#define	HUGEVAL			9999
#define	isedge( i )		(((i)==0)||((i)==7)||((i)==56)||((i)==63))
#define	isnedge( i )	(((i)==1)||((i)==6)||((i)==8)||((i)==15)||((i)==48)||((i)==55)||((i)==59)||((i)==62))
#define	CLRCHG(a)		(3-(a))

static
void xCopy( char* dst, char* src )
{
	int i;

	for ( i = 0; i < 64; i++ )
		dst[i] = src[i];
}

static
int sakiyomi_step0( char* board, int color, int max,
                    int* p_place, int passflag, int nhand )
{
	char	bd[64];
	int		i;
	int		place;
	int		point;
	int		gain;
	int		dummy;
	int		cur;

	if ( nhand == 0 )
	{
		point = 0;

		for ( i = 0; i < 64; i++ )
		{
			if ( Put( board, i, color, NOT_CHANGE ) > 0 )
			{
				point--;
			}
		}

		return point * 4;
	}

	place = -1;
	point = HUGEVAL;

	for ( i = 0; i < 64; i++ )
	{
		if ( ( gain = Put( board, i, color, NOT_CHANGE ) ) > 0 )
		{
			gain = gain * 2 + 1;
			if ( isedge( i ) )
				gain += 64;
			xCopy( bd, board );
			Put( bd, i, color, CHANGE );
			cur = gain - sakiyomi_step0( bd, CLRCHG( color ), gain - point, &dummy, 0, nhand-1 );
			if ( cur < point )
			{
				if ( cur <= max )
					return cur;
				place = i;
				point = cur;
			}
		}
	}

	if ( place == -1 )
	{
		if ( passflag )
			return 0;
		point = - sakiyomi_step0( board, CLRCHG( color ), - point, &dummy, 1, nhand-1 );
	}

	*p_place = place;
	return point;
}

static
int sakiyomi_step1( char* board, int color, int max,
                    int* p_place, int passflag, int nhand )
{
	char	bd[64];
	int		i;
	int		place;
	int		point;
	int		gain;
	int		dummy;
	int		cur;

	if ( nhand == 0 )
	{
		point = 0;

		for ( i = 0; i < 64; i++ )
		{
			if ( Put( board, i, color, NOT_CHANGE ) > 0 )
			{
				point--;
			}
		}

		return point * 2;
	}

	place = -1;
	point = HUGEVAL;

	for ( i = 0; i < 64; i++ )
	{
		if ( ( gain = Put( board, i, color, NOT_CHANGE ) ) > 0 )
		{
			gain = gain * 2 + 1;
			if ( isedge( i ) )
				gain += 64;
			xCopy( bd, board );
			Put( bd, i, color, CHANGE );
			cur = gain - sakiyomi_step1( bd, CLRCHG( color ), gain - point, &dummy, 0, nhand-1 );
			if ( cur < point )
			{
				if ( cur <= max )
					return cur;
				place = i;
				point = cur;
			}
		}
	}

	if ( place == -1 )
	{
		if ( passflag )
			return 0;
		point = - sakiyomi_step1( board, CLRCHG( color ), - point, &dummy, 1, nhand-1 );
	}

	*p_place = place;
	return point;
}

static
int sakiyomi_step2( char* board, int color, int max,
                    int* p_place, int passflag, int nhand )
{
	char	bd[64];
	int		i;
	int		place;
	int		point;
	int		gain;
	int		dummy;
	int		cur;

	if ( nhand == 0 )
	{
		point = HUGEVAL;

		for ( i = 0; i < 64; i++ )
		{
			if ( ( gain = Put( board, i, color, NOT_CHANGE ) ) > 0 )
			{
				if ( isedge( i ) )
					gain += 32;
				if ( gain < point )
					point = gain;
			}
		}
		if ( point == HUGEVAL )
			return 0;

		return point * 2 + 1;
	}

	place = -1;
	point = HUGEVAL;

	for ( i = 0; i < 64; i++ )
	{
		if ( ( gain = Put( board, i, color, NOT_CHANGE ) ) > 0 )
		{
			gain = gain * 2 + 1;
			if ( isedge( i ) )
				gain += 64;
			xCopy( bd, board );
			Put( bd, i, color, CHANGE );
			cur = gain - sakiyomi_step2( bd, CLRCHG( color ), gain - point, &dummy, 0, nhand-1 );
			if ( cur < point )
			{
				if ( cur <= max )
					return cur;
				place = i;
				point = cur;
			}
		}
	}

	if ( place == -1 )
	{
		if ( passflag )
			return 0;
		point = - sakiyomi_step2( board, CLRCHG( color ), - point, &dummy, 1, nhand-1 );
	}

	*p_place = place;
	return point;
}

static
int sakiyomi_all( char* board, int color, int max,
                  int* p_place, int passflag )
{
	char	bd[64];
	int		i;
	int		place;
	int		point;
	int		gain;
	int		dummy;
	int		cur;

	place = -1;
	point = HUGEVAL;

	for ( i = 0; i < 64; i++ )
	{
		if ( ( gain = Put( board, i, color, NOT_CHANGE ) ) > 0 )
		{
			gain = gain * 2 + 1;
			xCopy( bd, board );
			Put( bd, i, color, CHANGE );
			cur = gain - sakiyomi_all( bd, CLRCHG( color ), gain - point, &dummy, 0 );
			if ( cur < point )
			{
				if ( cur <= max )
					return cur;
				place = i;
				point = cur;
			}
		}
	}

	if ( place == -1 )
	{
		if ( passflag )
			return 0;
		point = - sakiyomi_all( board, CLRCHG( color ), - point, &dummy, 1 );
	}

	*p_place = place;
	return point;
}

static
int space_count( char* board )
{
	int	i,j;

	CountDisks( board, &i, &j );
	return ( 64 - ( i + j ) );
}

int takeshim2(char *board, int status)
{
	int place;
	int space;

	if ( passp(status) || initp(status) )
		return 0;

	space = space_count( board );

	if ( space == 60 )
		return 19;
	else if ( space >= 53 )
		sakiyomi_step0( board, MY_COLOR, - HUGEVAL, &place, 0, 3 + (space&1) );
	else if ( space >= 49 )
		sakiyomi_step0( board, MY_COLOR, - HUGEVAL, &place, 0, 5 + (space&1) );
	else if ( space >= 26 )
		sakiyomi_step1( board, MY_COLOR, - HUGEVAL, &place, 0, 3 );
	else if ( space >= 20 )
		sakiyomi_step1( board, MY_COLOR, - HUGEVAL, &place, 0, 5 );
	else if ( space >= 16 )
		sakiyomi_step2( board, MY_COLOR, - HUGEVAL, &place, 0, 5 );
	else if ( space >= 12 )
		sakiyomi_step2( board, MY_COLOR, - HUGEVAL, &place, 0, 7 );
	else
		sakiyomi_all( board, MY_COLOR, - HUGEVAL, &place, 0 );

	return place;
}
